@@ -26,10 +26,12 @@
 #include <functional>
 #include <iomanip>
 #include <iostream>
+#include <iterator>
 #include <memory>
 #include <stdexcept>
 #include <vector>
 
+#include <limits.h>
 #include <termios.h>
 #include <unicode/putil.h>
 #include <unicode/uclean.h>
@@ -108,8 +110,9 @@ namespace {
 class SuppressEcho {
  public:
   explicit SuppressEcho() {
-    if (tcgetattr(STDIN_FILENO, &tattr_) < 0)
+    if (tcgetattr(STDIN_FILENO, &tattr_) < 0) {
       return;
+    }
 
     reset_ = true;
 
@@ -119,8 +122,9 @@ class SuppressEcho {
   }
 
   ~SuppressEcho() {
-    if (reset_)
+    if (reset_) {
       tcsetattr(STDIN_FILENO, TCSAFLUSH, &tattr_);
+    }
   }
 
   explicit operator bool() const { return reset_; }
@@ -173,30 +177,38 @@ bool Tree::ReadPasswordFromStdIn() {
 
   {
     const SuppressEcho guard;
-    if (guard)
+    if (guard) {
       std::cout << "Password > " << std::flush;
+    }
 
     // Read password from standard input.
-    if (!std::getline(std::cin, password))
+    if (!std::getline(std::cin, password)) {
       password.clear();
+    }
 
-    if (guard)
+    if (guard) {
       std::cout << "Got it!" << std::endl;
+    }
   }
 
   // Remove newline at the end of password.
-  while (password.ends_with('\n'))
+  while (password.ends_with('\n')) {
     password.pop_back();
+  }
 
   if (password.empty()) {
-    Log(LOG_DEBUG, "Got an empty password");
+    LOG(DEBUG) << "Got an empty password";
     return false;
   }
 
-  Log(LOG_DEBUG, "Got a password of ", password.size(), " bytes");
+  LOG(DEBUG) << "Got a password of " << password.size() << " bytes";
 
-  if (zip_set_default_password(zip_, password.c_str()) < 0)
-    throw ZipError("Cannot set password", zip_);
+  for (const auto& [zip, zip_path] : zips_) {
+    zip_t* const z = zip.get();
+    if (zip_set_default_password(z, password.c_str()) < 0) {
+      throw ZipError(StrCat("Cannot set password for ", Path(zip_path)), z);
+    }
+  }
 
   return true;
 }
@@ -213,9 +225,10 @@ class IcuGuard {
     udata_setCommonData(mapped_file_.data(), &error);
     udata_setFileAccess(UDATA_ONLY_PACKAGES, &error);
     u_init(&error);
-    if (U_FAILURE(error))
+    if (U_FAILURE(error)) {
       throw std::runtime_error(
           StrCat("Cannot initialize ICU: ", u_errorName(error)));
+    }
   }
 
   // Cleans up the ICU library.
@@ -258,7 +271,7 @@ class ConverterToUtf8 {
                                         in.data(), in.size(), &error);
 
     if (U_FAILURE(error)) {
-      Log(LOG_ERR, "Cannot convert to UTF-16: ", u_errorName(error));
+      LOG(ERROR) << "Cannot convert to UTF-16: " << u_errorName(error);
       return {};
     }
 
@@ -266,7 +279,7 @@ class ConverterToUtf8 {
                                          utf16.data(), len16, &error);
 
     if (U_FAILURE(error)) {
-      Log(LOG_ERR, "Cannot convert to UTF-8: ", u_errorName(error));
+      LOG(ERROR) << "Cannot convert to UTF-8: " << u_errorName(error);
       return {};
     }
 
@@ -281,9 +294,10 @@ class ConverterToUtf8 {
     UErrorCode error = U_ZERO_ERROR;
     ConverterPtr conv(ucnv_open(encoding, &error));
 
-    if (U_FAILURE(error))
+    if (U_FAILURE(error)) {
       throw std::runtime_error(StrCat("Cannot open converter for encoding '",
                                       encoding, "': ", u_errorName(error)));
+    }
 
     assert(conv);
     return conv;
@@ -306,12 +320,12 @@ std::string DetectEncoding(const std::string_view bytes) {
   const UCharsetMatch* const ucm = ucsdet_detect(csd.get(), &error);
   const char* const encoding = ucsdet_getName(ucm, &error);
   if (U_FAILURE(error)) {
-    Log(LOG_ERR, "Cannot detect encoding: ", u_errorName(error));
+    LOG(ERROR) << "Cannot detect encoding: " << u_errorName(error);
     return std::string();
   }
 
-  Log(LOG_DEBUG, "Detected encoding ", encoding, " with ",
-      ucsdet_getConfidence(ucm, &error), "% confidence");
+  LOG(DEBUG) << "Detected encoding " << encoding << " with "
+             << ucsdet_getConfidence(ucm, &error) << "% confidence";
 
   // Check if we want to convert the detected encoding via ICU.
   const std::string_view candidates[] = {
@@ -319,8 +333,9 @@ std::string DetectEncoding(const std::string_view bytes) {
       "ISO-2022-CN", "ISO-2022-JP", "ISO-2022-KR", "KOI8-R"};
 
   for (const std::string_view candidate : candidates) {
-    if (candidate == encoding)
+    if (candidate == encoding) {
       return encoding;
+    }
   }
 
   // Not handled by ICU.
@@ -331,37 +346,34 @@ std::string DetectEncoding(const std::string_view bytes) {
 
 Tree::~Tree() {
 #ifndef NDEBUG
-  files_by_original_path_.clear();
-
-  for (FileNode& node : files_by_path_)
+  for (FileNode& node : files_by_path_) {
     node.children.clear();
+  }
 #endif
 
   files_by_path_.clear_and_dispose(std::default_delete<FileNode>());
-
-  if (zip_close(zip_) != 0)
-    Log(LOG_ERR, "Error while closing archive: ", zip_strerror(zip_));
 }
 
-size_t Tree::GetBucketCount(zip_t* zip) {
-  const size_t n = std::min<zip_int64_t>(zip_get_num_entries(zip, 0),
-                                         std::numeric_limits<ssize_t>::max());
+size_t Tree::GetBucketCount(const Zips& zips) {
+  i64 n = 0;
+  for (const auto& [z, _] : zips) {
+    n += zip_get_num_entries(z.get(), 0);
+  }
   // Floor the number of elements to a power of 2 with a minimum of 16.
-  return std::bit_floor(n | 16u);
+  return std::bit_floor(static_cast<size_t>(
+      std::clamp<i64>(n, 16, std::numeric_limits<ssize_t>::max())));
 }
 
-void Tree::BuildTree() {
-  const zip_int64_t n = zip_get_num_entries(zip_, 0);
-
-  FileNode::Ptr root(new FileNode{
-      .zip = zip_, .data = {.nlink = 2, .mode = S_IFDIR | 0755}, .name = "/"});
-  assert(!root->parent);
-  [[maybe_unused]] const auto [pos, ok] = files_by_path_.insert(*root);
+Tree::Tree(std::span<const std::string> paths, Options opts)
+    : Tree(OpenZips(paths), std::move(opts)) {
+  // Register root node.
+  assert(root_);
+  assert(!root_->parent);
+  bool const ok = files_by_path_.insert(*root_).second;
   assert(ok);
-  root.release();  // Now owned by |files_by_path_|.
 
   // Sum of all uncompressed file sizes.
-  uint64_t total_uncompressed_size = 0;
+  i64 total_uncompressed_size = 0;
   zip_stat_t sb;
 
   // Concatenate all the names in a buffer in order to guess the encoding.
@@ -369,269 +381,301 @@ void Tree::BuildTree() {
   all_names.reserve(10000);
   size_t max_name_length = 0;
 
-  // search for absolute or parent-relative paths
-  for (zip_int64_t id = 0; id < n; ++id) {
-    if (zip_stat_index(zip_, id, ZIP_FL_ENC_RAW, &sb) < 0)
-      throw ZipError(StrCat("Cannot read entry #", id), zip_);
+  for (const auto& [zip, zip_path] : zips_) {
+    zip_t* const z = zip.get();
+    const i64 n = zip_get_num_entries(z, 0);
+    for (i64 id = 0; id < n; ++id) {
+      if (zip_stat_index(z, id, ZIP_FL_ENC_RAW, &sb) < 0) {
+        throw ZipError(
+            StrCat("Cannot read entry #", id, " of ", Path(zip_path)), z);
+      }
 
-    if ((sb.valid & ZIP_STAT_SIZE) != 0)
-      total_uncompressed_size += sb.size;
+      if ((sb.valid & ZIP_STAT_SIZE) != 0) {
+        total_uncompressed_size += sb.size;
+      }
 
-    if ((sb.valid & ZIP_STAT_NAME) == 0 || !sb.name || !*sb.name)
-      continue;
+      if ((sb.valid & ZIP_STAT_NAME) == 0 || !sb.name || !*sb.name) {
+        continue;
+      }
 
-    const std::string_view name = sb.name;
-    if (max_name_length < name.size())
-      max_name_length = name.size();
+      const std::string_view name = sb.name;
+      if (max_name_length < name.size()) {
+        max_name_length = name.size();
+      }
 
-    if (all_names.size() + name.size() <= all_names.capacity())
-      all_names.append(name);
-
-    if (!need_prefix_)
-      need_prefix_ = name.starts_with('/') || name.starts_with("../");
+      if (all_names.size() + name.size() <= all_names.capacity()) {
+        all_names.append(name);
+      }
+    }
   }
 
-  Log(LOG_DEBUG, "Total uncompressed size = ", total_uncompressed_size,
-      " bytes");
+  LOG(DEBUG) << "Total uncompressed size = " << total_uncompressed_size
+             << " bytes";
 
   // Detect filename encoding.
   std::string encoding;
-  if (opts_.encoding)
+  if (opts_.encoding) {
     encoding = opts_.encoding;
-  if (encoding.empty() || encoding == "auto")
+  }
+
+  if (encoding.empty() || encoding == "auto") {
     encoding = DetectEncoding(all_names);
+  }
+
   all_names.clear();
 
   // Prepare functor to convert filenames to UTF-8.
   // By default, just rely on the conversion to UTF-8 provided by libzip.
-  std::function<std::string_view(std::string_view)> toUtf8 =
-      [](std::string_view s) { return s; };
+  ToUtf8 toUtf8 = [](std::string_view s) { return s; };
   zip_flags_t zipFlags = ZIP_FL_ENC_GUESS;
 
   // But if the filename encoding is one of the encodings we want to convert
   // using ICU, prepare and use the ICU converter.
   if (!encoding.empty() && encoding != "libzip") {
     try {
-      if (encoding != "raw")
+      if (encoding != "raw") {
         toUtf8 = [converter = std::make_shared<ConverterToUtf8>(
                       encoding.c_str(), max_name_length)](std::string_view s) {
           return (*converter)(s);
         };
+      }
       zipFlags = ZIP_FL_ENC_RAW;
     } catch (const std::exception& e) {
-      Log(LOG_ERR, e.what());
+      LOG(ERROR) << e.what();
     }
   }
 
-  struct Hardlink {
-    zip_int64_t id;
+  struct HardLink {
+    zip_t* zip;
+    std::string prefix;
+    i64 id;
     mode_t mode;
   };
 
-  std::vector<Hardlink> hardlinks;
-  std::string path;
+  std::vector<HardLink> hard_links;
   Beat should_display_progress;
-  uint64_t total_extracted_size = 0;
+  i64 total_extracted_size = 0;
   const auto progress = [&should_display_progress, &total_uncompressed_size,
                          &total_extracted_size](const ssize_t chunk_size) {
     assert(chunk_size >= 0);
     total_extracted_size += chunk_size;
-    if (!should_display_progress)
-      return;
-    Log(LOG_INFO, "Loading ",
-        total_extracted_size < total_uncompressed_size
-            ? 100 * total_extracted_size / total_uncompressed_size
-            : 100,
-        "%");
+    if (should_display_progress) {
+      LOG(INFO) << "Loading "
+                << (total_extracted_size < total_uncompressed_size
+                        ? 100 * total_extracted_size / total_uncompressed_size
+                        : 100)
+                << "%";
+    }
   };
 
-  // Add zip entries for all items except hardlinks
-  for (zip_int64_t id = 0; id < n; ++id) {
-    if (zip_stat_index(zip_, id, zipFlags, &sb) < 0)
-      throw ZipError(StrCat("Cannot read entry #", id), zip_);
+  std::string path;
+  path.reserve(PATH_MAX);
 
-    const Path original_path =
-        (sb.valid & ZIP_STAT_NAME) != 0 && sb.name && *sb.name ? sb.name : "-";
-    const uint64_t size = (sb.valid & ZIP_STAT_SIZE) != 0 ? sb.size : 0;
-    const auto [mode, is_hardlink] = GetEntryAttributes(id, original_path);
-    const FileType type = GetFileType(mode);
-
-    const Path original_path_utf8 = toUtf8(original_path);
-    if (!Path::Normalize(&path, original_path_utf8, need_prefix_)) {
-      Log(LOG_ERR, "Skipped ", type, " [", id, "]: Cannot normalize path ",
-          original_path_utf8);
-      assert(total_uncompressed_size >= size);
-      total_uncompressed_size -= size;
-      continue;
+  // Add zip entries for all items except hard links
+  for (const auto& [zip, zip_path] : zips_) {
+    path = '/';
+    if (!opts_.merge) {
+      path += Path(zip_path).Split().second.WithoutExtension();
     }
+    const std::string prefix = path;
+    size_t const initial_path_length = path.size();
 
-    if (type == FileType::Directory) {
-      FileNode* const node = CreateDir(path);
-      assert(node->link == &node->data);
-      const nlink_t nlink = node->data.nlink;
-      assert(nlink >= 2);
-      node->data = DataNode::Make(zip_, id, mode);
-      node->data.nlink = nlink;
-      node->original_path = Path(original_path).WithoutTrailingSeparator();
-      files_by_original_path_.insert(*node);
+    zip_t* const z = zip.get();
+    const i64 n = zip_get_num_entries(z, 0);
+    for (i64 id = 0; id < n; ++id) {
+      if (zip_stat_index(z, id, zipFlags, &sb) < 0) {
+        throw ZipError(
+            StrCat("Cannot read entry #", id, " of ", Path(zip_path)), z);
+      }
+
+      const Path original_path =
+          (sb.valid & ZIP_STAT_NAME) != 0 && sb.name && *sb.name ? sb.name
+                                                                 : "-";
+      path.resize(initial_path_length);
+      Path(toUtf8(original_path)).NormalizeAppend(&path);
+      const i64 size = (sb.valid & ZIP_STAT_SIZE) != 0 ? sb.size : 0;
+      const auto [mode, is_hard_link] =
+          GetEntryAttributes(z, id, original_path);
+      const FileType type = GetFileType(mode);
+
+      if (type == FileType::Directory) {
+        FileNode* const node = CreateDir(path);
+        assert(node);
+        assert(!node->link);
+        const ino_t ino = node->data.ino;
+        const nlink_t nlink = node->data.nlink;
+        assert(nlink >= 2);
+        node->data = DataNode::Make(z, id, mode);
+        node->data.ino = ino;
+        node->data.nlink = nlink;
+        total_block_count_ += 1;
+        assert(total_uncompressed_size >= size);
+        total_uncompressed_size -= size;
+        continue;
+      }
+
+      if (type != FileType::File &&
+          (type == FileType::Symlink ? !opts_.include_symlinks
+                                     : !opts_.include_special_files)) {
+        LOG(INFO) << "Skipped " << type << " [" << ++DataNode::ino_count << "] "
+                  << Path(path);
+        assert(total_uncompressed_size >= size);
+        total_uncompressed_size -= size;
+        continue;
+      }
+
+      if (is_hard_link) {
+        if (opts_.include_hard_links) {
+          hard_links.push_back({z, prefix, id, mode});
+        } else {
+          LOG(INFO) << "Skipped " << type << " [" << ++DataNode::ino_count
+                    << "] " << Path(path);
+        }
+        assert(total_uncompressed_size >= size);
+        total_uncompressed_size -= size;
+        continue;
+      }
+
+      const auto [parent_path, name] = Path(path).Split();
+      FileNode* const parent = CreateDir(parent_path);
+      FileNode* const node = CreateFile(z, id, parent, name, mode);
+      assert(node->parent == parent);
+      parent->AddChild(node);
       total_block_count_ += 1;
-      assert(total_uncompressed_size >= size);
-      total_uncompressed_size -= size;
-      continue;
-    }
+      total_block_count_ += node->operator DataNode::Stat().st_blocks;
 
-    if (type != FileType::File &&
-        (type == FileType::Symlink ? !opts_.include_symlinks
-                                   : !opts_.include_special_files)) {
-      Log(LOG_INFO, "Skipped ", type, " [", id, "] ", Path(path));
-      assert(total_uncompressed_size >= size);
-      total_uncompressed_size -= size;
-      continue;
-    }
-
-    if (is_hardlink) {
-      if (opts_.include_hardlinks) {
-        hardlinks.push_back({id, mode});
-      } else {
-        Log(LOG_INFO, "Skipped ", type, " [", id, "] ", Path(path));
-      }
-      assert(total_uncompressed_size >= size);
-      total_uncompressed_size -= size;
-      continue;
-    }
-
-    const auto [parent_path, name] = Path(path).Split();
-    FileNode* const parent = CreateDir(parent_path);
-    FileNode* const node = CreateFile(id, parent, name, mode);
-    assert(node->parent == parent);
-    parent->AddChild(node);
-    node->original_path = original_path;
-    files_by_original_path_.insert(*node);
-    total_block_count_ += 1;
-    total_block_count_ += node->operator DataNode::Stat().st_blocks;
-
-    if (!zip_encryption_method_supported(sb.encryption_method, 1)) {
-      ZipError e(StrCat("Cannot decrypt ", *node, ": ",
-                        EncryptionMethod(sb.encryption_method)),
-                 ZIP_ER_ENCRNOTSUPP);
-      if (opts_.check_compression)
-        throw std::move(e);
-      Log(LOG_ERR, e.what());
-    }
-
-    if (!zip_compression_method_supported(sb.comp_method, 1)) {
-      ZipError e(StrCat("Cannot decompress ", *node, ": ",
-                        CompressionMethod(sb.comp_method)),
-                 ZIP_ER_COMPNOTSUPP);
-      if (opts_.check_compression)
-        throw std::move(e);
-      Log(LOG_ERR, e.what());
-    }
-
-    // Check the password on encrypted files.
-    if ((sb.valid & ZIP_STAT_ENCRYPTION_METHOD) != 0 &&
-        sb.encryption_method != ZIP_EM_NONE) {
-      CheckPassword(node);
-    }
-
-    // Cache file data if necessary.
-    if (opts_.pre_cache) {
-      try {
-        if (!node->CacheAll(progress)) {
-          assert(total_uncompressed_size >= size);
-          total_uncompressed_size -= size;
+      if (!zip_encryption_method_supported(sb.encryption_method, 1)) {
+        ZipError e(StrCat("Cannot decrypt ", *node, ": ",
+                          EncryptionMethod(sb.encryption_method)),
+                   ZIP_ER_ENCRNOTSUPP);
+        if (opts_.check_compression) {
+          throw std::move(e);
         }
-      } catch (const ZipError& error) {
-        Log(LOG_ERR, "Cannot cache ", *node, ": ", error.what());
-        if (opts_.check_password) {
-          Log(LOG_INFO,
-              "Use the --force option to continue even if some files cannot be "
-              "cached");
-          throw;
+        LOG(ERROR) << e.what();
+      }
+
+      if (!zip_compression_method_supported(sb.comp_method, 1)) {
+        ZipError e(StrCat("Cannot decompress ", *node, ": ",
+                          CompressionMethod(sb.comp_method)),
+                   ZIP_ER_COMPNOTSUPP);
+        if (opts_.check_compression) {
+          throw std::move(e);
+        }
+        LOG(ERROR) << e.what();
+      }
+
+      // Check the password on encrypted files.
+      if ((sb.valid & ZIP_STAT_ENCRYPTION_METHOD) != 0 &&
+          sb.encryption_method != ZIP_EM_NONE) {
+        CheckPassword(node);
+      }
+
+      // Cache file data if necessary.
+      if (opts_.pre_cache) {
+        try {
+          if (!node->CacheAll(progress)) {
+            assert(total_uncompressed_size >= size);
+            total_uncompressed_size -= size;
+          }
+        } catch (const ZipError& error) {
+          LOG(ERROR) << "Cannot cache " << *node << ": " << error.what();
+          if (opts_.check_password) {
+            LOG(INFO) << "Use the -o force option to continue even if some "
+                         "files cannot be cached";
+            throw;
+          }
         }
       }
     }
   }
 
-  // Add hardlinks
-  for (const auto [id, mode] : hardlinks) {
-    const Path original_path = zip_get_name(zip_, id, zipFlags);
-    const FileType type = GetFileType(mode);
-
-    const Path original_path_utf8 = toUtf8(original_path);
-    if (!Path::Normalize(&path, original_path_utf8, need_prefix_)) {
-      Log(LOG_ERR, "Skipped ", type, " [", id, "]: Cannot normalize path ",
-          original_path_utf8);
-      continue;
-    }
-
+  // Add hard links
+  for (auto& [z, prefix, id, mode] : hard_links) {
+    const Path original_path = zip_get_name(z, id, zipFlags);
+    path = prefix;
+    Path(toUtf8(original_path)).NormalizeAppend(&path);
     const auto [parent_path, name] = Path(path).Split();
     FileNode* const parent = CreateDir(parent_path);
-    FileNode* node = CreateHardlink(id, parent, name, mode);
+    assert(parent);
+    FileNode* const node =
+        CreateHardLink(z, id, parent, name, mode, std::move(prefix), toUtf8);
+    assert(node);
     assert(node->parent == parent);
     parent->AddChild(node);
-    node->original_path = original_path;
-    files_by_original_path_.insert(*node);
     total_block_count_ += 1;
   }
 
-  if (should_display_progress.Count())
-    Log(LOG_INFO, "Loaded 100%");
+  if (should_display_progress.Count()) {
+    LOG(INFO) << "Loaded 100%";
+  }
 
-  Log(LOG_DEBUG, "Nodes = ", GetNodeCount());
-  Log(LOG_DEBUG, "Blocks = ", total_block_count_);
+  // Trim the top level if necessary.
+  if (opts_.trim) {
+    assert(root_);
+    if (opts_.merge) {
+      Trim(*root_);
+    } else {
+      for (FileNode& c : root_->children) {
+        Trim(c);
+      }
+    }
+  }
+
+  LOG(DEBUG) << "Nodes = " << GetNodeCount();
+  LOG(DEBUG) << "Blocks = " << total_block_count_;
 }
 
 void Tree::CheckPassword(const FileNode* const node) {
   assert(node);
 
-  if (checked_password_)
+  if (checked_password_) {
     return;
+  }
 
-  Log(LOG_INFO, "Need password for ", *node);
+  LOG(INFO) << "Need password for " << *node;
   ReadPasswordFromStdIn();
 
   try {
-    Log(LOG_DEBUG, "Checking password on ", *node, "...");
+    LOG(DEBUG) << "Checking password on " << *node << "...";
 
     // Try to open the file and read a few bytes from it.
-    const ZipFile file(zip_fopen_index(zip_, node->id, 0));
-    if (!file)
-      throw ZipError(StrCat("Cannot open ", *node), zip_);
+    const ZipFile file(zip_fopen_index(node->zip, node->id, 0));
+    if (!file) {
+      throw ZipError(StrCat("Cannot open ", *node), node->zip);
+    }
 
     std::array<char, 16> buf;
-    if (zip_fread(file.get(), buf.data(), buf.size()) < 0)
+    if (zip_fread(file.get(), buf.data(), buf.size()) < 0) {
       throw ZipError(StrCat("Cannot read ", *node), file.get());
+    }
 
-    Log(LOG_INFO, "Password is Ok");
+    LOG(INFO) << "Password is Ok";
   } catch (const ZipError& error) {
     if (opts_.check_password) {
-      Log(LOG_INFO,
-          "Use the --force option to mount an encrypted ZIP with a wrong "
-          "password");
+      LOG(INFO) << "Use the -o force option to mount an encrypted ZIP with a "
+                   "wrong password";
       throw;
     }
 
-    Log(LOG_DEBUG, error.what());
-    Log(LOG_INFO,
-        "Continuing despite wrong password because of --force option");
+    LOG(DEBUG) << error.what();
+    LOG(INFO) << "Continuing despite wrong password because of -o force option";
   }
 
   checked_password_ = true;
 }
 
 Tree::EntryAttributes Tree::GetEntryAttributes(
-    const zip_uint64_t id,
+    zip_t* const z,
+    const i64 id,
     const std::string_view original_path) {
   const bool is_dir = original_path.ends_with('/');
 
   zip_uint8_t opsys;
   zip_uint32_t attr;
-  zip_file_get_external_attributes(zip_, id, 0, &opsys, &attr);
+  zip_file_get_external_attributes(z, id, 0, &opsys, &attr);
 
-  mode_t unix_mode = attr >> 16;
-  mode_t mode;
-  bool is_hardlink = false;
+  mode_t mode = static_cast<mode_t>(attr >> 16);
+  bool is_hard_link = false;
 
   /*
    * PKWARE describes "OS made by" now (since 1998) as follows:
@@ -640,48 +684,41 @@ Tree::EntryAttributes Tree::GetEntryAttributes(
    * and can be read by PKZIP for DOS version 2.04g then this value will be
    * zero.
    */
-  if (opsys == ZIP_OPSYS_DOS && GetFileType(unix_mode) != FileType::Unknown)
+  if (opsys == ZIP_OPSYS_DOS && GetFileType(mode) != FileType::Unknown) {
     opsys = ZIP_OPSYS_UNIX;
+  }
 
   const zip_uint32_t FZ_ATTR_HARDLINK = 0x800;
 
   switch (opsys) {
     case ZIP_OPSYS_UNIX:
-      mode = unix_mode;
-
       // force is_dir value
       if (is_dir) {
         SetFileType(&mode, FileType::Directory);
-      } else {
-        switch (GetFileType(mode)) {
-          case FileType::Unknown:    // treat unknown file types as regular
-          case FileType::Directory:  // mislabeled as directory
-            SetFileType(&mode, FileType::File);
-            break;
-          default:
-            break;
-        }
+      } else if (const FileType ft = GetFileType(mode);
+                 ft == FileType::Unknown || ft == FileType::Directory) {
+        // If unknown type or mislabeled as directory, relabel as regular file.
+        SetFileType(&mode, FileType::File);
       }
 
-      // Always ignore hardlink flag for dirs
-      is_hardlink = (attr & FZ_ATTR_HARDLINK) != 0 && !is_dir;
+      // Always ignore hard link flag for dirs
+      is_hard_link = (attr & FZ_ATTR_HARDLINK) != 0 && !is_dir;
       break;
 
     case ZIP_OPSYS_DOS:
     case ZIP_OPSYS_WINDOWS_NTFS:
     case ZIP_OPSYS_MVS:
-      /*
-       * Both WINDOWS_NTFS and OPSYS_MVS used here because of
-       * difference in constant assignment by PKWARE and Info-ZIP
-       */
+      // Both WINDOWS_NTFS and OPSYS_MVS used here because of difference in
+      // constant assignment by PKWARE and Info-ZIP.
       mode = 0444;
+
       // http://msdn.microsoft.com/en-us/library/windows/desktop/gg258117%28v=vs.85%29.aspx
       // http://en.wikipedia.org/wiki/File_Allocation_Table#attributes
       // FILE_ATTRIBUTE_READONLY
       if ((attr & 1) == 0) {
         mode |= 0220;
       }
-      // directory
+
       if (is_dir) {
         mode |= S_IFDIR | 0111;
       } else {
@@ -698,17 +735,18 @@ Tree::EntryAttributes Tree::GetEntryAttributes(
       }
   }
 
-  return {mode, is_hardlink};
+  return {mode, is_hard_link};
 }
 
 FileNode* Tree::Attach(FileNode::Ptr node) {
   assert(node);
   const auto [pos, ok] = files_by_path_.insert(*node);
-  if (ok)
+  if (ok) {
     return node.release();  // Now owned by |files_by_path_|.
+  }
 
   // There is a name collision
-  Log(LOG_DEBUG, *node, " conflicts with ", *pos);
+  LOG(DEBUG) << *node << " conflicts with " << *pos;
 
   // Extract filename extension
   std::string& f = node->name;
@@ -727,54 +765,59 @@ FileNode* Tree::Attach(FileNode::Ptr node) {
 
     const auto [pos, ok] = files_by_path_.insert(*node);
     if (ok) {
-      Log(LOG_DEBUG, "Resolved conflict for ", *node);
+      LOG(DEBUG) << "Resolved conflict for " << *node;
       return node.release();  // Now owned by |files_by_path_|.
     }
 
-    Log(LOG_DEBUG, *node, " conflicts with ", *pos);
-    if (!i)
+    LOG(DEBUG) << *node << " conflicts with " << *pos;
+    if (!i) {
       i = &pos->collision_count;
+    }
   }
 }
 
-FileNode* Tree::CreateFile(zip_int64_t id,
+FileNode* Tree::CreateFile(zip_t* const z,
+                           i64 id,
                            FileNode* parent,
                            std::string_view name,
                            mode_t mode) {
   assert(parent);
   assert(!name.empty());
   assert(id >= 0);
-  return Attach(
-      FileNode::Ptr(new FileNode{.zip = zip_,
-                                 .id = id,
-                                 .data = DataNode::Make(zip_, id, mode),
-                                 .parent = parent,
-                                 .name = std::string(name)}));
+  return Attach(FileNode::Ptr(new FileNode{.zip = z,
+                                           .id = id,
+                                           .data = DataNode::Make(z, id, mode),
+                                           .parent = parent,
+                                           .name = std::string(name)}));
 }
 
-FileNode* Tree::CreateHardlink(zip_int64_t id,
+FileNode* Tree::CreateHardLink(zip_t* const z,
+                               i64 id,
                                FileNode* parent,
                                std::string_view name,
-                               mode_t mode) {
+                               mode_t mode,
+                               std::string target_path,
+                               const ToUtf8& toUtf8) {
   assert(parent);
   assert(!name.empty());
   assert(id >= 0);
 
   FileNode::Ptr node(new FileNode{
-      .zip = zip_, .id = id, .parent = parent, .name = std::string(name)});
+      .zip = z, .id = id, .parent = parent, .name = std::string(name)});
 
   zip_uint16_t len;
   const zip_uint8_t* field = zip_file_extra_field_get_by_id(
-      zip_, id, FZ_EF_PKWARE_UNIX, 0, &len, ZIP_FL_CENTRAL);
-
-  if (!field)
-    field = zip_file_extra_field_get_by_id(zip_, id, FZ_EF_PKWARE_UNIX, 0, &len,
-                                           ZIP_FL_LOCAL);
+      z, id, FZ_EF_PKWARE_UNIX, 0, &len, ZIP_FL_CENTRAL);
 
   if (!field) {
-    // Ignoring hardlink without PKWARE UNIX field
-    Log(LOG_INFO, "Cannot find PkWare Unix field for hardlink ", *node);
-    return CreateFile(id, parent, name, mode);
+    field = zip_file_extra_field_get_by_id(z, id, FZ_EF_PKWARE_UNIX, 0, &len,
+                                           ZIP_FL_LOCAL);
+  }
+
+  if (!field) {
+    // Ignoring hard link without PKWARE UNIX field
+    LOG(INFO) << "Cannot find PkWare Unix field for hard link " << *node;
+    return CreateFile(z, id, parent, name, mode);
   }
 
   time_t mt, at;
@@ -786,39 +829,40 @@ FileNode* Tree::CreateHardlink(zip_int64_t id,
 
   if (!ExtraField::parsePkWareUnixField(len, field, mode, mt, at, uid, gid, dev,
                                         link, link_len)) {
-    Log(LOG_WARNING, "Cannot parse PkWare Unix field for hardlink ", *node);
-    return CreateFile(id, parent, name, mode);
+    LOG(WARNING) << "Cannot parse PkWare Unix field for hard link " << *node;
+    return CreateFile(z, id, parent, name, mode);
   }
 
   if (link_len == 0 || !link) {
-    Log(LOG_ERR, "Cannot get target for hardlink ", *node);
-    return CreateFile(id, parent, name, mode);
+    LOG(ERROR) << "Cannot get target for hard link " << *node;
+    return CreateFile(z, id, parent, name, mode);
   }
 
-  const std::string_view target_path(link, link_len);
+  Path(toUtf8(Path(link, link_len))).NormalizeAppend(&target_path);
 
-  const auto it = files_by_original_path_.find(
-      Path(target_path).WithoutTrailingSeparator());
-  if (it == files_by_original_path_.end()) {
-    Log(LOG_ERR, "Cannot find target for hardlink ", *node, " -> ",
-        Path(target_path));
-    return CreateFile(id, parent, name, mode);
+  const auto it = files_by_path_.find(target_path);
+  if (it == files_by_path_.end()) {
+    LOG(ERROR) << "Cannot find target for hard link " << *node << " -> "
+               << Path(target_path);
+    return CreateFile(z, id, parent, name, mode);
   }
 
   const FileNode& target = *it;
 
-  if (target.type() != GetFileType(mode)) {
+  if (target.GetType() != GetFileType(mode)) {
     // PkZip saves hard-link flag for symlinks with inode link count > 1.
-    if (!S_ISLNK(mode))
-      Log(LOG_ERR, "Mismatched types for hardlink ", *node, " -> ", target);
+    if (!S_ISLNK(mode)) {
+      LOG(ERROR) << "Mismatched types for hard link " << *node << " -> "
+                 << target;
+    }
 
-    return CreateFile(id, parent, name, mode);
+    return CreateFile(z, id, parent, name, mode);
   }
 
-  node->link = target.link;
+  node->link = &target.GetTarget();
   node->link->nlink++;
 
-  Log(LOG_DEBUG, "Created hardlink ", *node, " -> ", target);
+  LOG(DEBUG) << "Created hard link " << *node << " -> " << target;
   return Attach(std::move(node));
 }
 
@@ -835,13 +879,14 @@ FileNode* Tree::CreateDir(std::string_view path) {
   FileNode* parent;
 
   if (FileNode* const node = Find(path)) {
-    if (node->is_dir())
+    if (node->IsDir()) {
       return node;
+    }
 
     // There is an existing node with the given name, but it's not a
     // directory.
-    Log(LOG_DEBUG, "Found conflicting ", *node, " while creating Dir ",
-        Path(path));
+    LOG(DEBUG) << "Found conflicting " << *node << " while creating Dir "
+               << Path(path);
     parent = node->parent;
 
     // Remove it from |files_by_path_|, in order to insert it again later with a
@@ -853,31 +898,95 @@ FileNode* Tree::CreateDir(std::string_view path) {
   }
 
   assert(parent);
-  FileNode::Ptr child(new FileNode{.zip = zip_,
-                                   .data = {.nlink = 2, .mode = S_IFDIR | 0755},
+  FileNode::Ptr child(new FileNode{.data = {.nlink = 2, .mode = S_IFDIR | 0755},
                                    .parent = parent,
                                    .name = std::string(name)});
-  assert(child->path() == path);
+  assert(child->GetPath() == path);
   parent->AddChild(child.get());
   [[maybe_unused]] const auto [pos, ok] = files_by_path_.insert(*child);
   assert(ok);
-  parent->link->nlink++;
+  FileNode* const ret = child.release();  // Now owned by |files_by_path_|.
+  assert(!parent->link);
+  parent->data.nlink++;
 
-  if (to_rename)
+  if (to_rename) {
     Attach(std::move(to_rename));
+  }
 
-  return child.release();  // Now owned by |files_by_path_|.
+  return ret;
 }
 
-Tree::Ptr Tree::Init(const char* const filename, Options opts) {
-  assert(filename);
-  int err;
-  zip_t* const zip_file = zip_open(filename, ZIP_RDONLY, &err);
+Tree::Zips Tree::OpenZips(std::span<const std::string> paths) {
+  Zips zips;
+  zips.reserve(paths.size());
 
-  if (!zip_file)
-    throw ZipError(StrCat("Cannot open ZIP archive ", Path(filename)), err);
+  for (const std::string& path : paths) {
+    int err;
+    Zip z(zip_open(path.c_str(), ZIP_RDONLY, &err));
 
-  Ptr tree(new Tree(zip_file, std::move(opts)));
-  tree->BuildTree();
-  return tree;
+    if (!z) {
+      throw ZipError(StrCat("Cannot open ZIP archive ", Path(path)), err);
+    }
+
+    LOG(DEBUG) << "Opened " << Path(path);
+    zips.push_back({std::move(z), path});
+  }
+
+  return zips;
+}
+
+void Tree::Trim(FileNode& a) {
+  FileNode* p = a.GetUniqueChildDirectory();
+  if (!p) {
+    return;
+  }
+
+  Deindex(*p);
+  a.children.clear();
+
+  while (FileNode* const q = p->GetUniqueChildDirectory()) {
+    p->children.clear();
+    p = q;
+  }
+
+  LOG(DEBUG) << "Collapsing " << *p << " into " << a;
+
+  a.data = std::move(p->data);
+  assert(!a.link);
+  assert(!p->link);
+  a.children = std::move(p->children);
+  assert(p->children.empty());
+
+  for (FileNode& c : a.children) {
+    assert(c.parent == p);
+    c.parent = &a;
+    Reindex(c);
+  }
+
+  LOG(INFO) << "Collapsed " << *p << " into " << a;
+  LOG(INFO)
+      << "Use `-o notrim` if you want to keep these intermediate directories";
+
+  while (p != &a) {
+    total_block_count_ -= 1;
+    FileNode* const q = p->parent;
+    delete p;
+    p = q;
+  }
+}
+
+void Tree::Deindex(FileNode& node) {
+  for (FileNode& c : node.children) {
+    Deindex(c);
+  }
+  files_by_path_.erase(files_by_path_.iterator_to(node));
+}
+
+void Tree::Reindex(FileNode& node) {
+  bool const ok = files_by_path_.insert(node).second;
+  assert(ok);
+
+  for (FileNode& c : node.children) {
+    Reindex(c);
+  }
 }
